/*
    libporty - async dns resolving
    Copyright (C) 2010..2012  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page: http://repo.hu/projects/libporty
    Author: libporty@igor2.repo.hu
*/

#include "adns.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <libporty/bgtask/bgtask.h>
#include <libporty/net/network.h>
#include <libporty/net/dns4.h>
#include <libporty/net/local.h>
#include <libporty/util/buf.h>
#include <libporty/util/argv.h>

static void P_dns_action(int fd, int argc, char *argv[])
{
	char *cmd, *protocol, *data, *cookie;
	char wb[1024];
	char name[32];
	int len;

	/* make rest of the code more readable */
	cmd      = argv[0];
	protocol = argv[1];
	data     = argv[2];
	cookie   = argv[3];

	/* snprintf is not portable - cut protocol type to something reasonable */
	if (strlen(cookie) > 128)
		 cookie[128] = '\0';

	if (strcmp(protocol, "IPv4") != 0) {
		/* snprintf is not portable - cut protocol type to something reasonable */
		if (strlen(protocol) > 32)
			 protocol[32] = '\0';
		len = sprintf(wb, "E 1 %s protocol '%s' is not supported\n", cookie, protocol);
		send(fd, wb, len, 0);
		return;
	}

	if (strcmp(cmd, "addr") == 0) {
		P_ipv4_addr_t ip;

		if ((P_dns4_name_to_IP(&ip, data) != 0) || (P_dns4_IP_to_addr(&ip, name) != 0))
			len = sprintf(wb, "E 4 %s resolving error\n", cookie);
		else
			len = sprintf(wb, "D addr %s %s\n", name, cookie);
		send(fd, wb, len, 0);
	}
	else if (strcmp(cmd, "name") == 0) {
		char *name = NULL;
		P_ipv4_addr_t ip;

		P_dns4_addr_to_IP(&ip, data);
		P_dns4_IP_to_addr(&ip, wb);

		if ((P_dns4_addr_to_IP(&ip, data) != 0) || (P_dns4_IP_to_name(&name, &ip) != 0))
			len = sprintf(wb, "E 4 %s resolving error\n", cookie);
		else
			len = sprintf(wb, "D name %s %s\n", name, cookie);
		send(fd, wb, len, 0);
		if (name != NULL)
			free(name);
	}
	else {
		len = sprintf(wb, "E 3 %s unknown command '%s'\n", cookie, cmd);
		send(fd, wb, len, 0);
	}
}

static int P_dns_thread(void *arg)
{
	P_net_socket *fd = (P_net_socket *)arg;
	P_buf_t *buf;
	int len;
	char rb[256];
	char *request;
	struct P_pollfd fds[1];
	char *argv[4];
	int argc;

	fds[0].fd = *fd;
	fds[0].events = P_POLLIN;

	buf = P_buf_init("\r\n");
	for(;;) {
		P_poll(fds, 1, 1000);
		if (fds[0].revents & P_POLLIN) {
			len = recv(*fd, rb, sizeof(rb)-1, 0);
			if (len == 0)
				goto quit;
			if (len > 0) {
				rb[len] = '\0';
				P_buf_fill(buf, rb);
				while((request = P_buf_get(buf)) != NULL) {
					P_argv_split(request, &argc, argv, 4, " \t");
					if ((argc > 0) && (*argv[0] == 'q')) {
						/* assume 'quit' and kill the task */
						P_buf_free(request);
						goto quit;
					}
					if (argc < 3) {
						len = sprintf(rb, "E 2 - not enough arguments\n");
						send(*fd, rb, len, 0);
					}
					else
						P_dns_action(*fd, argc, argv);
					P_buf_free(request);
				}
			}
		}
	}

quit:;
	P_net_close(*fd);
	P_buf_uninit(buf);
	return 0;

}

int P_net_async_dns(P_net_socket *fd)
{
	int ret;
	P_net_socket *chfd;

	chfd = calloc(sizeof(P_net_socket), 1);
	ret = P_net_local_socket(fd, chfd, P_net_nonblock_full);
	if (ret != 0)
		return ret;

	P_bgtask_create(P_dns_thread, chfd);

	return 0;
}


/* ----------------------- helper functions ------------------------------------ */

int P_net_async_dns_req(P_net_socket fd, P_net_async_dns_req_t *req)
{
	char line[1024], *payload, *s, *o;
	int l, n;

	/* no request */
	if ((req->hostname == NULL) && (req->ip_addr == NULL))
		return -1;

	/* request already fulfilled */
	if ((req->hostname != NULL) && (req->ip_addr != NULL))
		return -2;

	/* check if payload is too long */
	payload = req->hostname == NULL ? req->ip_addr : req->hostname;
	if (strlen(payload) > sizeof(line) - (16 + sizeof(req->cookie.s) * 2))
		return -3;

	/* check if payload would break protocol */
	for(s = payload; *s != '\0'; s++) {
		if ((*s < 33) || (*s > 126))
			return -4;
	}

	/* assamble command part of the request */
	l = sprintf(line, "%s IPv4 %s ", req->hostname == NULL ? "name" : "addr", payload);

	/* convert cookie to text */
	o = line+l;
	for(s = req->cookie.s, n = 0; n < sizeof(req->cookie.s); n++, s++) {
		*o = ((*s >> 4) & 0x0F) + 'A';
		o++;
		*o = (*s & 0x0F) + 'A';
		o++;
	}
	*o = '\n';
	o++;

	send(fd, line, o-line, 0);
	return 0;
}

int P_net_async_dns_read(void **buf_, P_net_socket fd)
{
	P_buf_t **buf = (P_buf_t **)buf_;
	char in[256];
	int len;

	len = recv(fd, in, sizeof(in)-1, 0);
	/* on eof or error, destroy buffer (if we have one) and return error */
	if (len <= 0) {
		if (*buf != NULL) {
			P_buf_uninit(*buf);
			*buf = NULL;
		}
		return -1;
	}

	/* first call, create buffer */
	if (*buf == NULL)
		*buf = P_buf_init("\r\n");

	in[len] = '\0';

	P_buf_fill(*buf, in);
	return 0;
}

static char *strclone(const char *str)
{
	int len;
	char *s;

	len = strlen(str);
	s   = malloc(len+1);
	memcpy(s, str, len+1);
	return s;
}

int P_net_async_dns_parse(void **buf_, P_net_async_dns_req_t *ans)
{
	P_buf_t **buf = (P_buf_t **)buf_;
	char *cmd, *i, *o;
	char *argv[5];
	int argc, n;

	cmd = P_buf_get(*buf);
	if (cmd == NULL)
		return -1;

	P_argv_split(cmd, &argc, argv, 5, " \t");
	if (argc < 2)
		goto error_internal;
	switch(*argv[0]) {
		case 'D':
			if (argc < 4)
				goto error_internal;
			if (strcmp(argv[1], "name") == 0)
				ans->hostname = strclone(argv[2]);
			else if (strcmp(argv[1], "addr") == 0)
				ans->ip_addr = strclone(argv[2]);
			else
				goto error_internal;

			for(i = argv[3], o = ans->cookie.s, n = 0; n < sizeof(ans->cookie.s); n++, o++) {
				if (*i == '\0')
					goto error_internal;
				*o = ((*i - 'A') << 4) & 0xF0;
				i++;
				if (*i == '\0')
					goto error_internal;
				*o |= (*i - 'A') & 0x0F;
				i++;
			}

			ans->outcome = 0;
			goto quit;
		case 'E':
			ans->outcome = atoi(argv[1]);
			goto quit;
	}

error_internal:;
	ans->outcome = P_NET_DNS_OUT_INTERNAL;
quit:;
	P_buf_free(cmd);
	return 0;
}

int P_net_async_dns_destroy(void **buf_, P_net_socket fd)
{
	P_buf_t **buf = (P_buf_t **)buf_;

	if (*buf != NULL) {
		P_buf_uninit(*buf);
		*buf = NULL;
	}

	send(fd, "quit\n", 5, 0);
	P_net_close(fd);
	return 0;
}
