/*
    c99tree - scope and typedef handling
    Copyright (C) 2010-2015  Tibor Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "c99tree-cc.h"
#include "scope.h"
#include "genht/hash.h"

/* --- hash --- */
static int keyeq(char *a, char *b)
{
	return !strcmp(a, b);
}

static unsigned int hash(char *key)
{
	return murmurhash(key, strlen(key));
}

/* --- scope --- */
void c99c_scope_enter(c99tree_cc_t *ctx)
{
	c99c_scope_t *s;

	s = malloc(sizeof(c99c_scope_t));

	s->typedefs = htsp_alloc(hash, keyeq);
	s->negative_typedefs = htsp_alloc(hash, keyeq);
	s->compound = NULL;

	s->child = NULL;
	s->parent = ctx->scope;
	if (ctx->scope != NULL)
		ctx->scope->child = s;
	ctx->scope = s;
}

void c99c_scope_leave(c99tree_cc_t *ctx)
{
	htsp_entry_t *e;
	c99c_scope_t *s = ctx->scope;

	assert(ctx->scope->child == NULL);
	ctx->scope = s->parent;
	if (ctx->scope != NULL)
		ctx->scope->child = NULL;
	for (e = htsp_first(s->typedefs); e; e = htsp_next(s->typedefs, e)) {
		free(e->key);
		C99Nfree(ctx, e->value);
	}
	for (e = htsp_first(s->negative_typedefs); e; e = htsp_next(s->negative_typedefs, e)) {
		free(e->key);
		C99Nfree(ctx, e->value);
	}
	htsp_free(s->typedefs);
	htsp_free(s->negative_typedefs);
	free(s);
}

/* --- typedef --- */

void c99c_scope_add_typedef(c99tree_cc_t *ctx, const char *old, c99c_node_t *new)
{
	htsp_set(ctx->scope->typedefs, (htsp_key_t)c99c_strclone(old), (htsp_value_t)new);
}


const c99c_node_t *c99c_scope_get_typedef(c99tree_cc_t *ctx, const char *type)
{
	c99c_scope_t *s;
	const c99c_node_t *ret;

	/* walk trough all parents as well, return first hit... */
	for(s = ctx->scope; s != NULL; s = s->parent) {
		/* if we have a negative typedef, that overrides any later real typedef */
		ret = (const c99c_node_t *)htsp_get(s->negative_typedefs, (htsp_key_t)type);
		if (ret != NULL)
			return NULL;

		/* return the first typedef found in any of the current or parent scopes */
		ret = (const c99c_node_t *)htsp_get(s->typedefs, (htsp_key_t)type);
		if (ret != NULL)
			return ret;
	}

	/* ...or return NULL if nothing is found (not ever typedef'd) */
	return NULL;
}

void c99c_scope_typedef_add(c99tree_cc_t *ctx, c99c_node_t *type, c99c_node_t *ids)
{
	c99c_node_t *n;

	assert(type->type == T_TYPE);

	for(n = ids; n != NULL; n = n->next) {
		assert(n->type == T_ID);
		c99c_scope_add_typedef(ctx, n->name, C99Nalloc(ctx, T_TYPECUSTOM, c99c_strclone(n->name), ctx->verbose_options[VO_TYPEDEF] ? C99Nclone(ctx, type, INV_LOC_NONE) : NULL));
	}
}

void c99c_scope_negative_typedef_add(c99tree_cc_t *ctx, const char *name)
{
	htsp_set(ctx->scope->negative_typedefs, (htsp_key_t)c99c_strclone(name), C99Nalloc(ctx, T_INVALID, NULL, NULL));
}
